(**************************************************************************)
(*                                                                        *)
(*                               GlSurf                                   *)
(*                                                                        *)
(*                   C. Raffalli, Universite de Savoie                    *)
(*                                                                        *)
(* Copyright 2003, 2004 Christohe Raffalli                                *)
(*                                                                        *)
(*  This file is part of GlSurf.                                          *)
(*                                                                        *)
(*  GlSurf is free software; you can redistribute it and/or modify        *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation; either version 2 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  GlSurf is distributed in the hope that it will be useful,             *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *) 
(*  You should have received a copy of the GNU General Public License     *)
(*  along with GlSurf; if not, write to the Free Software                 *)
(*  Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *)
(*  MA  02111-1307  USA                                                   *)
(**************************************************************************)

open Algebra
open Format 

module StringMap : Map.S with type key = string

(* this is a subtype of both Field and Vectorial_Space *)
module type Expression  =
sig
  (* include Field *)
  type scalar

  type trans = (* transantes function, like sin() or cos() *)
      {
	(* the name *)
	name : string;
	(* the function that computes the value *)
	calcul : scalar -> scalar;
        (* the derivate *)
	deriv : elem;
	(* the name of the variable you used to write the derivate *)
	name_var: string
      }
      
  and trans2 = (* transantes function, like sin() or cos() *)
      {
	name2 : string;
	calcul2 : scalar -> scalar -> scalar;
	deriv1 : elem;
	deriv2 : elem;
	name_var1: string;
	name_var2: string
      }

  and user_fun  = (* user defined functions *)
      {
	fname : string;
        variables : string list;
        value : elem
      }

  (* the internal type used to represent expressions *)
  and elem =
      Add of elem  * elem
    | Sub of elem  * elem
    | Div of elem  * elem
    | Mul of elem  * elem
    | Pow of elem  * int
    | Tra of int * elem
    | Tra2 of int * elem * elem
    | Cst of scalar
    | Var of string

  (* environment used to store values of variables *)
  type 'a environment = (string * 'a) list

(*** field struct ***)
  val zero : elem              (* zero ! *)
  val one :  elem              (* one ! *)
  val t_of_int : int -> elem   (* the standard mapping from integer *)
  val (++) : elem -> elem -> elem         
  val (--) : elem -> elem -> elem       
  val ( ** ) : elem -> elem -> elem      
  val ( @ ) : scalar -> elem -> elem      
  val (==) : elem -> elem -> bool      
  val opp : elem -> elem             
  val inv : elem -> elem
  val (//) : elem -> elem -> elem    
  val cst : scalar -> elem
  val var : string -> elem

  val trans_table : int StringMap.t ref
  val trans_array : trans array ref
  val trans2_table : int StringMap.t ref
  val trans2_array : trans2 array ref
  val fun_table : user_fun StringMap.t ref
  val loc_var : string list ref

  type share
  type ishare
  val print_mode : [`C | `Normal | `C_Shared of share | `C_Interval of ishare] ref
  val print : elem -> unit          
  val write : formatter -> elem -> unit
  val write_interval : formatter -> elem -> unit
  val write_to_string : elem -> string
  val parse : char Stream.t -> elem
  val read : in_channel -> elem    
  val write_bin : out_channel -> elem -> unit
  val read_bin : in_channel -> elem 

  val init_share : unit -> share
  val build_share : share -> elem -> unit
  val write_share : formatter -> share -> unit

  val init_ishare : string list -> ishare
  val build_ishare : ishare -> elem -> unit
  val write_ishare : formatter -> ishare -> unit

  val simplify : elem -> elem
  val develop : elem -> elem
  val normalize : elem -> elem
  val derive   : elem -> string -> elem
  (* derive n times *)
  val laplacien : elem -> string list -> elem				     
  val sder     : elem -> string -> int -> elem
  val eval     : elem -> scalar environment -> scalar	
  val subst    : elem -> elem environment -> elem
  val subst'    : elem -> scalar environment -> elem
  (* conjugate all constant in the expression, live the variable unchanged !*)
  val conjugate : elem -> elem
  val varlist : elem -> string list

  val decompose_poly : elem -> (scalar * (string * int) list) list
  val make_poly : elem -> elem 
end                                                                                       

module Expression (R:Field) : Expression
  with type scalar = R.elem

