(**************************************************************************)
(*                                                                        *)
(*                               GlSurf                                   *)
(*                                                                        *)
(*                   C. Raffalli, Universite de Savoie                    *)
(*                                                                        *)
(* Copyright 2003, 2004 Christohe Raffalli                                *)
(*                                                                        *)
(*  This file is part of GlSurf.                                          *)
(*                                                                        *)
(*  GlSurf is free software; you can redistribute it and/or modify        *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation; either version 2 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  GlSurf is distributed in the hope that it will be useful,             *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *) 
(*  You should have received a copy of the GNU General Public License     *)
(*  along with GlSurf; if not, write to the Free Software                 *)
(*  Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *)
(*  MA  02111-1307  USA                                                   *)
(**************************************************************************)

open Format
open Algebra 

module type Optimisation_Space =
sig
  type scalar
  type elem
  module Scalar : Normed_Field with type elem = scalar and type norm = scalar
  module Space : Banach with 
    type scalar = scalar and type norm = scalar and type elem = elem

  type precomputation 

  exception Exit_domain

  val precompute : elem -> precomputation
  val functional : precomputation -> scalar
  val differential : scalar -> precomputation -> elem
  val invhessian : scalar -> precomputation -> elem -> elem

  (* side effect called at each step of optimisation *)
  val change : elem * precomputation -> unit

  (* called each time we advance some step, may be the identity, 
  simplification ... or a projection for constrained optimisation *)
  val project : elem -> elem 
end



type cg_kind = 
  | Fletcher_Reeves
  | Polak_Ribiere
  | Hestenes_Stiefel
  | Dai_Yuan
      
type algorithm =
  | Gradient
  | Conjugate_gradient of cg_kind * int
  | AvGrad
  | Newton of algorithm

val parse_algorithm : char Stream.t -> algorithm

type ('a) parameters =
    {
     initial_lambda : 'a;
     stop_condition : 'a;
     algorithm : algorithm;
     linear_search_precision : 'a;
     min_lambda : 'a;
     max_lambda : 'a;
    log : formatter option;
   }

module type Atype =
sig
  type elem
end

module Linear_optimisation :
  functor (Atype : Atype) ->
  functor (Scalar : Normed_Field with type elem = Atype.elem and type norm = Atype.elem) ->
  sig
    val linear_optimisation :
        ('a -> 'b) ->
          (Scalar.elem -> 'a * Scalar.norm) ->
            Scalar.elem ->
		Scalar.elem ->
		  Scalar.norm ->
		    Scalar.elem -> Scalar.norm -> 'a -> Scalar.elem * Scalar.norm * 'a
  end

module type Descent =
  sig
    type scalar
    type elem

    val default : (scalar) parameters
    val optimise : elem -> (scalar) parameters -> elem * scalar
  end

module Descent : functor (O : Optimisation_Space) ->
  Descent with  
    type scalar = O.Scalar.elem and
    type elem = O.Space.elem

